/*
 *	ipddp.h: Header for IP-over-DDP driver for Linux.
 */

#ifndef __LINUX_IPDDP_H
#define __LINUX_IPDDP_H

#ifdef __KERNEL__

#define SIOCADDIPDDPRT SIOCDEVPRIVATE
#define SIOCDELIPDDPRT SIOCDEVPRIVATE+1
#define SIOCFINDIPDDPRT SIOCDEVPRIVATE+2

struct ipddp_route
{
        struct device *dev;             /* Carrier device */
        __u32 ip;                       /* IP address */
        struct at_addr at;              /* Gateway appletalk address */
        int flags;
        struct ipddp_route *next;
};

static struct ipddp_route *ipddp_route_head;
static struct ipddp_route ipddp_route_test;

#endif	/* __KERNEL__ */
#endif	/* __LINUX_IPDDP_H */
