/*
 * SNI specific definitions
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1997 by Ralf Baechle
 */
#ifndef __ASM_MIPS_DESKSTATION_H 
#define __ASM_MIPS_DESKSTATION_H 

#define RPC44_PORT_BASE	0xe2000000

#endif /* __ASM_MIPS_DESKSTATION_H */
